import time
import networkx as nx
from gurobipy import *

""" 
    Source code of POP model in 
    "A new Integer Linear Program for the Steiner Tree Problem with Revenues, Budget and Hop Constraints"


    author: Adalat Jabrayilov
    adalat.jabrayilov@tu-dortmund.de
    Di 14. Aug 15:39:01 CEST 2018
"""


def hstp( G,Root,PV,B,H,PRINT=0,TLIM=1000,IP="pop2" ): #{
    """ 
    (POP) 
    =====
     + (EURC)
     + (basic) + (deltaRoot) + (pi>len) + (nonterminal-innode) + (terminal-leaf)

    PRINT is 3 Bit Number x2x1x0 \in { 0,..,7 }
	x0 :  Gurobi Output	# 1 
	x1 :  layout chanel   	# 2
	x2 :  debug  chanel 1 	# 4 
	x3 :  debug  chanel 2 	# 8 

    """
    c00 = time.clock()
    
    """ remove reflexive edges (v,v) \in E """
    for u,v in G.edges():
	if u == v:  G.remove_edge(u,v)

    """ zero fill """
    for v in G.nodes():
        if v not in PV:
            PV[v] = 0

    V_0 = G.nodes()
    E_0 = G.edges()
                                                  
    if IP == "pop1" or IP == "pop2":
        dist_Root = nx.shortest_path_length( G, source=Root, weight=None )
    elif IP == "pop1r" or IP == "pop2r":
        #   _ __  _ __ ___ _ __  _ __ ___   ___ ___  ___ ___ 
        #  | '_ \| '__/ _ \ '_ \| '__/ _ \ / __/ _ \/ __/ __|
        #  | |_) | | |  __/ |_) | | | (_) | (_|  __/\__ \__ \
        #  | .__/|_|  \___| .__/|_|  \___/ \___\___||___/___/
        #  |_|            |_|                                
        G, dist_Root = preprocessing( G, Root, PV, H, PRINT)
    else:
        print "unknown ILP:", IP
        return 2


    V  = G.nodes()
    E  = G.edges()
    Vr = [ v for v in V if v != Root ]

    rt_red=time.clock() - c00; 
    if PRINT&4: print 'Reduction (V,E): (%s,%s) --> (%s,%s)    %0.2F sec  ' % (
		len(V_0), len(E_0), len(V), len(E), rt_red ); 

    #  __  __           _      _ 
    # |  \/  | ___   __| | ___| |
    # | |\/| |/ _ \ / _` |/ _ \ |
    # | |  | | (_) | (_| |  __/ |
    # |_|  |_|\___/ \__,_|\___|_|
    #                            
    p = Model("hstp")
    p.params.OutputFlag = 1&PRINT
    p.params.SEED       = 1	    # randomized=False
    p.params.THREADS    = 1	    # nur ein thread
    p.params.TimeLimit  = TLIM
    p.params.mipgap = 0.00001

    
    # POP variables
    x = {}
    for u,v in E:
        x[u,v] = p.addVar(vtype=GRB.BINARY) 
        x[v,u] = p.addVar(vtype=GRB.BINARY) 

    y = { (i,u): p.addVar(vtype=GRB.BINARY) for u in V for i in range(H+1) }

    p.update()


    if PRINT&8:	print 'Root:', Root


    y [0,Root] = 0
    for v in Vr:    #{
        y[ 0, v ].LB = 1
        y[ H, v ].UB = 0

        """ (*pi>len) """
        if dist_Root[v] <= H:
            y[ dist_Root[v]-1, v ].LB = 1

        """ (nonterminal-innode) """
        if PV[ v ] == 0 and dist_Root[v] <= H-1:
            y[ H-1, v ].UB = 0
    #}

    p.update()

    """    _     _           _   _       
      ___ | |__ (_) ___  ___| |_(_)_   __
     / _ \| '_ \| |/ _ \/ __| __| \ \ / /
    | (_) | |_) | |  __/ (__| |_| |\ V / 
     \___/|_.__// |\___|\___|\__|_| \_/  
              |__/                       
    """
    p.setObjective( 
	PV[Root] + sum ([ PV[v]*x[u,v] + PV[u]*x[v,u] for u,v in E ])
	, GRB.MAXIMIZE
    )

    """        _     _           _   
     ___ _   _| |__ (_) ___  ___| |_ 
    / __| | | | '_ \| |/ _ \/ __| __|
    \__ \ |_| | |_) | |  __/ (__| |_ 
    |___/\__,_|_.__// |\___|\___|\__|
		  |__/               
    """
    # (3,4)
    for u in V:
	for i in range(H): 
	    p.addConstr( y[i,u] - y[i+1,u] >= 0 ) # (3)

    # ( Partial Ordering )
    for u,v in E:
	# i = 0
	p.addConstr( y[0,v] - x[u,v] >= 0 )  #(5)
	p.addConstr( y[0,u] - x[v,u] >= 0 )  #(5)
	for i in range(1,H+1):    
	    p.addConstr( 1-y[i-1,u] + y[i,v] - x[u,v] >= 0 )   # (5)
	    p.addConstr( 1-y[i-1,v] + y[i,u] - x[v,u] >= 0 )   # (6)


    # (indeg+outdeg)
    for v in Vr:
        # (*indeg)
        p.addConstr(               sum([ x[u,v] for u in G.neighbors(v) ]) <= 1 )    #(9)
        # (*outdeg) 
        for w in G.neighbors(v):
            p.addConstr( x[v,w] <= sum([ x[u,v] for u in G.neighbors(v) if u != w ])      )    #(9)

    # (*delta(Root))
    for v in G.neighbors(Root): 
        if H > 0 and G[Root][v]['weight'] <= B:
            p.addConstr( sum([ x[Root,v] for v in G.neighbors(Root) if G[Root][v]['weight'] <= B ])  >= 1 ) 
            break

    # (*B)
    p.addConstr( sum([ d['weight']*(x[u,v]+x[v,u]) for u,v,d in G.edges(data=True) ]) <= B )    # (10)

    if IP == "pop2" or IP == "pop2r":
        """ (terminal-leaf) """
        for v in Vr:
            if PV[ v ] > 0:
                p.addConstr( sum([ x[v,w] for w in G.neighbors(v) ]) >= 1 - y[H-1,v] )


    rt_write   = (time.clock() - c00)

    p.optimize()

    rt_solv = p.runtime 

    rt_total = rt_red + rt_write + rt_solv 

    rt_red   = '%.2F' % rt_red
    rt_write = '%.2F' % rt_write
    rt_solv  = '%.2F' % rt_solv
    rt_total = '%.2F' % rt_total

    print 'H', H
    print 'B', B
    print len(V_0), len(E_0), len(V), len(E), p.objVal, p.objBound, rt_red, rt_write, rt_solv, rt_total,
#}


#                                                   _             
#  _ __  _ __ ___ _ __  _ __ ___   ___ ___  ___ ___(_)_ __   __ _ 
# | '_ \| '__/ _ \ '_ \| '__/ _ \ / __/ _ \/ __/ __| | '_ \ / _` |
# | |_) | | |  __/ |_) | | | (_) | (_|  __/\__ \__ \ | | | | (_| |
# | .__/|_|  \___| .__/|_|  \___/ \___\___||___/___/_|_| |_|\__, |
# |_|            |_|                                        |___/ 

def preprocessing( G, Root, PV, H, PRINT ): #{
    """
    ALGORITHM:
    ==========
    1. EURC : extended undirected root cost test
    """ 

    if PRINT&8: print 'initial SIZE', len(G.edges())

    G = EURC( G,Root,PRINT )
    if PRINT&8: print 'EURC', len(G.edges())

    return G, nx.shortest_path_length( G, source=Root, weight=None )
#}

def path_cost( G, path=[] ): #{
    cost = 0
    for i in range(1, len(path)):
	u,v   = path[i-1:i+1]
	cost += G[u][v]['weight']
	#print '-', u,v,  G[u][v]['weight']
    return cost
#}
def path_dist_cost( G, pathRoot=dict() ): #{
    dist = {v:(len(pathRoot[v])-1) for v in G}
    cost = {v:path_cost(G,pathRoot[v]) for v in G}
    return dist, cost
#}

def EURC(G,Root,PRINT): #{
    """ 
        Extended undirected root-cost test.

        ALGORITHM:
        ==========
        it transform the undirected graph G to directed graph D.
        If both arcs (u,v) and (v,u) are deleted from D, then the 
        coresponding undirected edge uv is deleted in G.
    """

    D = nx.DiGraph()
    for u,v,d in G.edges( data = True ):
        D.add_edge( u,v,d )
        D.add_edge( v,u,d )
    for v in D.predecessors( Root ):
        D.remove_edge( v,Root )

    path_weighted = nx.shortest_path( D, source=Root, weight='weight' )
    P1_dist_Root, P1_cost_Root = path_dist_cost( D, path_weighted )

    path_unweighted = nx.shortest_path( D, source=Root, weight=None )
    P2_dist_Root, P2_cost_Root = path_dist_cost( D, path_unweighted )

    # unweighted shortest path length from Root
    dist_Root = P2_dist_Root

    ArcsToRemove=[]
    for u,v,d in D.edges( data=True ):
        if u != Root and v != Root:
            c_uv = d['weight']
            d_ru = dist_Root[u]

	    if  (  ( P1_cost_Root[v] <= c_uv ) and ( P1_dist_Root[v] <= d_ru +1 )
	        or ( P2_cost_Root[v] <= c_uv ) and ( P2_dist_Root[v] <= d_ru +1 ) ):
		ArcsToRemove.append( (u,v) )
    if PRINT&16: print 'rm', ArcsToRemove
    D.remove_edges_from( ArcsToRemove )

    G = nx.Graph()
    for u,v,d in D.edges( data=True ):
        G.add_edge( u,v,d )

    return G
#}
