import re, networkx

""" 
    Parser for the Steiner Tree Problem with Revenues, Budget and Hop Constraints


    author: Adalat Jabrayilov
    adalat.jabrayilov@tu-dortmund.de
    Di 14. Aug 15:57:28 CEST 2018
"""

class SteinLib_parser:
    """
        PARSER:
        =======
            it parses stprbh file, which is extention to SteinLib format.

        WARNING:
        ========
        This Parser assume:    
            - Graph is connected
        
        CODE_BASED_ON:
        ==============
            SteinLibParser.py

        RETURN:
        =======
            Typle (G,Root,PV,Budget,HopLimit) where:
                G : weighted graph where
                    E : list of Tuples ( src_id, trg_id, weight ) 
                PV = dict();   : dict of profitable vertices: key=v val=profit(v)

        Example:
        ========
	import stprbh_parser 
	G,Root,PV,B,H = stprbh_parser.SteinLib_parser().parse( "$graph" )
	print len(G.nodes())
	print len(G.edges())
	print Root
	print len(PV)
	print B
	print H
	for u,v,d in G.edges(data=True):
            w = d['weight']
	    print u,v,w

    """ 

    def __init__(self):
	self.G        = networkx.Graph(); """ Weighted graph """
	self.Root     = None
	self.PV       = dict();  """ PV: dict of profitable vertices: key=v val=profit(v) """
        self.Budget   = 0;
        self.HopLimit = 0;

    def text_of_section(self, name): #{
        """
        RETURN:
        =======
            Text of the section with name <name> where:
            <name> := Graph|ProfitableVertices|Coordinates
	
            re.S : FLAG for a newline, since a section contains >1 lines. 
	    (?i) : ignorecase
	"""
	section_pattern = re.compile('(?i)SECTION' +'\s+' +name +'.+' +'(?i)END',  re.S ); 
	section_obj = section_pattern.search ( self.text )
	if section_obj != None:
	    return section_obj.group(); 
	return None;
    #} 
    def section_Graph(self): #{ 
        text = self.text_of_section('Graph');

        # root
	self.Root     = re.search( 'Root\s+(\w+)', text ).group(1) 

        # Budget
	self.Budget   = int(re.search( 'Budget\s+(\w+)', text ).group(1)) 

        # HopLimit
	self.HopLimit = int(re.search( 'HopLimit\s+(\w+)', text ).group(1))

        # Edges
	edge_pattern = re.compile( 
	    '(E\s+(\w+)\s+(\w+)\s+(\w+))+'  # group 2,3,4
	);
	for item in edge_pattern.finditer( text ):
	    u = item.group(2) ;
	    v = item.group(3) ;
	    w = int( item.group(4) );
            self.G.add_edge (u, v, weight=w)
    #} 
    def section_ProfitableVertices(self): #{
	pv_pattern = re.compile( 
	    '(PV\s+(\w+)\s+(\w+))+'  # group 2
	); 
	for item in pv_pattern.finditer( self.text_of_section('ProfitableVertices') ):
	    vertex_id     = item.group(2);
	    vertex_profit = item.group(3);
	    self.PV[ vertex_id ] = int(vertex_profit);
    #} 
    def parse(self, stp_file): #{
	self.text = open( stp_file, 'r').read()
	self.section_Graph();
	self.section_ProfitableVertices();
	return self.G, self.Root, self.PV, self.Budget, self.HopLimit
    #} 
